# __init__.py/Open GoPro, Version 1.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue May 18 22:08:50 UTC 2021

"""All Protobuf defnitions."""

from open_gopro.proto.turbo_transfer_pb import RequestSetTurboActive
from open_gopro.proto.response_generic_pb import ResponseGeneric
from open_gopro.proto.request_get_preset_status_pb import RequestGetPresetStatus
from open_gopro.proto.preset_status_pb import NotifyPresetStatus

__all__ = ["RequestSetTurboActive", "ResponseGeneric", "RequestGetPresetStatus", "NotifyPresetStatus"]
