# request_get_preset_status_pb.py/Open GoPro, Version 1.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue May 18 22:08:50 UTC 2021

# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: request_get_preset_status.proto
# plugin: python-betterproto
from dataclasses import dataclass
from typing import List

import betterproto


class EnumRegisterPresetStatus(betterproto.Enum):
    REGISTER_PRESET_STATUS_PRESET = 1
    REGISTER_PRESET_STATUS_PRESET_GROUP_ARRAY = 2


@dataclass
class RequestGetPresetStatus(betterproto.Message):
    register_preset_status: List["EnumRegisterPresetStatus"] = betterproto.enum_field(1)
    unregister_preset_status: List["EnumRegisterPresetStatus"] = betterproto.enum_field(2)
