# __init__.py/Open GoPro, Version 1.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue May 18 22:08:50 UTC 2021

"""All GoPro parameters (both defined constants and generated from protobufs)."""

from open_gopro.proto.request_get_preset_status_pb import EnumRegisterPresetStatus
from open_gopro.params.params import *

__all__ = ["EnumRegisterPresetStatus"]
