# turbo_transfer_pb.py/Open GoPro, Version 1.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue May 18 22:08:50 UTC 2021

# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: turbo_transfer.proto
# plugin: python-betterproto
from dataclasses import dataclass

import betterproto


@dataclass
class RequestSetTurboActive(betterproto.Message):
    active: bool = betterproto.bool_field(1)
